@@ -42,6 +42,7 @@ Revision History:
 #include "sat/dimacs.h"
 #include "sat/sat_solver.h"
 #include "sat/tactic/goal2sat.h"
+#include "sat/tactic/sat2goal.h"
 
 
 extern "C" {
@@ -874,7 +875,7 @@ extern "C" {
         init_solver(c, s);
         solver::push_eh_t _push = push_eh;
         solver::pop_eh_t _pop = pop_eh;
-        solver::fresh_eh_t _fresh = [&](void * user_ctx, ast_manager& m, solver::context_obj*& _ctx) {
+        solver::fresh_eh_t _fresh = [=](void * user_ctx, ast_manager& m, solver::context_obj*& _ctx) {
             ast_context_params params;
             params.set_foreign_manager(&m);
             auto* ctx = alloc(api::context, &params, false);
