@@ -350,7 +350,7 @@ func_decl* seq_decl_plugin::mk_left_assoc_fun(decl_kind k, unsigned arity, sort*
     return mk_assoc_fun(k, arity, domain, range, k_seq, k_string, false);
 }
 
-func_decl* seq_decl_plugin::mk_ubv2s(unsigned arity, sort* const* domain) {
+func_decl* seq_decl_plugin::mk_ubv2s(unsigned arity, sort* const* domain) const {
     ast_manager& m = *m_manager;
     if (arity != 1)
         m.raise_exception("Invalid str.from_ubv expects one bit-vector argument");
@@ -361,6 +361,17 @@ func_decl* seq_decl_plugin::mk_ubv2s(unsigned arity, sort* const* domain) {
     return m.mk_func_decl(symbol("str.from_ubv"), arity, domain, rng, func_decl_info(m_family_id, OP_STRING_UBVTOS));    
 }
 
+func_decl* seq_decl_plugin::mk_sbv2s(unsigned arity, sort* const* domain) const {
+    ast_manager &m = *m_manager;
+    if (arity != 1)
+        m.raise_exception("Invalid str.from_sbv expects one bit-vector argument");
+    bv_util bv(m);
+    if (!bv.is_bv_sort(domain[0]))
+        m.raise_exception("Invalid str.from_sbv expects one bit-vector argument");
+    sort *rng = m_string;
+    return m.mk_func_decl(symbol("str.from_sbv"), arity, domain, rng, func_decl_info(m_family_id, OP_STRING_SBVTOS));
+}
+
 func_decl* seq_decl_plugin::mk_assoc_fun(decl_kind k, unsigned arity, sort* const* domain, sort* range, decl_kind k_seq, decl_kind k_string, bool is_right) {
     ast_manager& m = *m_manager;
     sort_ref rng(m);
@@ -376,7 +387,7 @@ func_decl* seq_decl_plugin::mk_assoc_fun(decl_kind k, unsigned arity, sort* cons
 }
 
 
-func_decl * seq_decl_plugin::mk_func_decl(decl_kind k, unsigned num_parameters, parameter const * parameters,
+func_decl* seq_decl_plugin::mk_func_decl(decl_kind k, unsigned num_parameters, parameter const * parameters,
                                           unsigned arity, sort * const * domain, sort * range) {
     init();
     m_has_seq = true;
@@ -416,8 +427,12 @@ func_decl * seq_decl_plugin::mk_func_decl(decl_kind k, unsigned num_parameters,
     case OP_STRING_FROM_CODE:
         match(*m_sigs[k], arity, domain, range, rng);
         return m.mk_func_decl(m_sigs[k]->m_name, arity, domain, rng, func_decl_info(m_family_id, k));
+
     case OP_STRING_UBVTOS:
-        return mk_ubv2s(arity, domain);        
+        return mk_ubv2s(arity, domain);
+
+    case OP_STRING_SBVTOS:
+        return mk_sbv2s(arity, domain);
 
     case _OP_REGEXP_FULL_CHAR:
         m_has_re = true;
@@ -627,6 +642,7 @@ void seq_decl_plugin::get_op_names(svector<builtin_name> & op_names, symbol cons
     op_names.push_back(builtin_name("re.nostr",  _OP_REGEXP_EMPTY));
     op_names.push_back(builtin_name("re.complement", OP_RE_COMPLEMENT));
     op_names.push_back(builtin_name("str.from_ubv", OP_STRING_UBVTOS));
+    op_names.push_back(builtin_name("str.from_sbv", OP_STRING_SBVTOS));
 }
 
 void seq_decl_plugin::get_sort_names(svector<builtin_name> & sort_names, symbol const & logic) {
@@ -782,6 +798,15 @@ bool seq_util::is_char2int(expr const* e) const {
     return ch.is_to_int(e);
 }
 
+bool seq_util::is_bv2char(expr const* e) const {
+    return ch.is_bv2char(e);
+}
+
+bool seq_util::is_char2bv(expr const* e) const {
+    return ch.is_char2bv(e);
+}
+
+
 app* seq_util::mk_char(unsigned ch) const {
     return seq.mk_char(ch);
 }
@@ -814,7 +839,7 @@ bool seq_util::str::is_nth_i(expr const* n, expr*& s, unsigned& idx) const {
     return arith_util(m).is_unsigned(i, idx);
 }
 
-app* seq_util::str::mk_nth_i(expr* s, unsigned i) const {
+app* seq_util::str::mk_nth_c(expr* s, unsigned i) const {
     return mk_nth_i(s, arith_util(m).mk_int(i));
 }
 
@@ -829,6 +854,48 @@ void seq_util::str::get_concat(expr* e, expr_ref_vector& es) const {
     }
 }
 
+/*
+Returns true if s is an expression of the form (l = |u|) |u|-k or (-k)+|u| or |u|+(-k).
+Also returns true and assigns k=0 and l=s if s is |u|.
+*/
+bool seq_util::str::is_len_sub(expr const* s, expr*& l, expr*& u, rational& k) const {
+    expr* x;
+    rational v;
+    arith_util a(m);
+    if (is_length(s, l)) {
+        k = 0;
+        return true;
+    }
+    else if (a.is_sub(s, l, x) && is_length(l, u) && a.is_numeral(x, v) && v.is_nonneg()) {
+        k = v;
+        return true;
+    }
+    else if (a.is_add(s, l, x) && is_length(l, u) && a.is_numeral(x, v) && v.is_nonpos()) {
+        k = - v;
+        return true;
+    }
+    else if (a.is_add(s, x, l) && is_length(l, u) && a.is_numeral(x, v) && v.is_nonpos()) {
+        k = - v;
+        return true;
+    }
+    else
+        return false;
+}
+
+bool seq_util::str::is_unit_string(expr const* s, expr_ref& c) const {
+    zstring z;
+    expr* ch = nullptr;
+    if (is_string(s, z) && z.length() == 1) {
+        c = mk_char(z[0]);
+        return true;
+    }
+    else if (is_unit(s, ch)) {
+        c = ch;
+        return true;
+    }
+    return false;
+}
+
 void seq_util::str::get_concat_units(expr* e, expr_ref_vector& es) const {
     expr* e1, *e2;
     while (is_concat(e, e1, e2)) {
@@ -851,8 +918,6 @@ app* seq_util::str::mk_is_empty(expr* s) const {
     return m.mk_eq(s, mk_empty(s->get_sort()));
 }
 
-
-
 unsigned seq_util::str::min_length(expr* s) const {
     SASSERT(u.is_seq(s));
     unsigned result = 0;
@@ -867,7 +932,10 @@ unsigned seq_util::str::min_length(expr* s) const {
             return 0u;
     };
     while (is_concat(s, s1, s2)) {
-        result += get_length(s1);
+        if (is_concat(s1))
+            result += min_length(s1);
+        else
+            result += get_length(s1);
         s = s2;
     }
     result += get_length(s);
@@ -895,7 +963,10 @@ unsigned seq_util::str::max_length(expr* s) const {
             return UINT_MAX;
     };
     while (is_concat(s, s1, s2)) {
-        result = u.max_plus(get_length(s), result);
+        if (is_concat(s1))
+            result = u.max_plus(max_length(s1), result);
+        else
+            result = u.max_plus(get_length(s1), result);
         s = s2;
     }
     result = u.max_plus(get_length(s), result);
@@ -1040,33 +1111,70 @@ app* seq_util::rex::mk_epsilon(sort* seq_sort) {
 /*
   Produces compact view of concrete concatenations such as (abcd).
 */
-std::ostream& seq_util::rex::pp::compact_helper_seq(std::ostream& out, expr* s) const {
-    SASSERT(re.u.is_seq(s));
+bool seq_util::rex::pp::print_seq(std::ostream& out, expr* s) const {
+    zstring z;
+    expr* x, * j, * k, * l, * i, * x_;
     if (re.u.str.is_empty(s))
         out << "()";
-    else if (re.u.str.is_unit(s))
-        seq_unit(out, s);
     else if (re.u.str.is_concat(s)) {
         expr_ref_vector es(re.m);
         re.u.str.get_concat(s, es);
         for (expr* e : es)
-            compact_helper_seq(out, e);
+            print(out, e);
     }
-    //using braces to indicate 'full' output
-    //for example an uninterpreted constant X will be printed as {X}
-    //while a unit sequence "X" will be printed as X
-    //thus for example (concat "X" "Y" Z "W") where Z is uninterpreted is printed as XY{Z}W
-    else out << "{" << mk_pp(s, re.m) << "}";
-    return out;
+    else if (re.u.str.is_string(s, z)) {
+        for (unsigned i = 0; i < z.length(); i++)
+            out << (char)z[i];
+    }
+    else if (re.u.str.is_at(s, x, i))
+        print(out, x) << "@", print(out, i);
+    else if (re.u.str.is_extract(s, x, j, k)) {
+        rational jv, iv;
+        print(out, x);
+        if (arith_util(re.m).is_numeral(j, jv)) {
+            if (arith_util(re.m).is_numeral(k, iv)) {
+                // output X[j,k]
+                out << "[" << jv.get_int32() << "," << jv.get_int32() << "]";
+            }
+            else if (arith_util(re.m).is_sub(k, l, i) && re.u.str.is_length(l, x_) && x == x_ &&
+                arith_util(re.m).is_numeral(i, iv) && iv == jv) {
+                // case X[j,|X|-j] is denoted by X[j..]
+                out << "[" << jv.get_int32() << "..]";
+            }
+            else if (((arith_util(re.m).is_add(k, l, i) && re.u.str.is_length(l, x_)) ||
+                (arith_util(re.m).is_add(k, i, l) && re.u.str.is_length(l, x_))) && x == x_ &&
+                arith_util(re.m).is_numeral(i, iv) && iv.get_int32() + jv.get_int32() == 0) {
+                // case X[j,|X|-j] is denoted by X[j..]
+                out << "[" << jv.get_int32() << "..]";
+            }
+            else {
+                out << "[" << jv.get_int32() << ",";
+                print(out, k);
+                out << "]";
+            }
+        }
+        else {
+            out << "[";
+            print(out, j);
+            out << ",";
+            print(out, k);
+            out << "]";
+        }
+    }
+    else
+        return false;
+    return true;
 }
 
 /*
   Produces output such as [a-z] for a range.
 */
-std::ostream& seq_util::rex::pp::compact_helper_range(std::ostream& out, expr* s1, expr* s2) const {
+std::ostream& seq_util::rex::pp::print_range(std::ostream& out, expr* s1, expr* s2) const {
     out << "[";
-    seq_unit(out, s1) << "-";
-    seq_unit(out, s2) << "]";
+    print(out, s1);
+    out << "-";
+    print(out, s2);
+    out << "]";
     return out;
 }
 
@@ -1081,10 +1189,11 @@ bool seq_util::rex::pp::can_skip_parenth(expr* r) const {
 /*
   Specialize output for a unit sequence converting to visible ASCII characters if possible.
 */
-std::ostream& seq_util::rex::pp::seq_unit(std::ostream& out, expr* s) const {
-    expr* e;
+bool seq_util::rex::pp::print_unit(std::ostream& out, expr* s) const {
+    expr* e, * i;
     unsigned n = 0;
-    if (re.u.str.is_unit(s, e) && re.u.is_const_char(e, n)) {
+
+    if ((re.u.str.is_unit(s, e) && re.u.is_const_char(e, n)) || re.u.is_const_char(s, n)) {
         char c = (char)n;
         if (c == '\n')
             out << "\\n";
@@ -1092,22 +1201,21 @@ std::ostream& seq_util::rex::pp::seq_unit(std::ostream& out, expr* s) const {
             out << "\\r";
         else if (c == '\f')
             out << "\\f";
-        else if (c == ' ')
-            out << "\\s";
-        else if (c == '(' || c == ')' || c == '{' || c == '}' || c == '[' || c == ']' || c == '.' || c == '\\')
-            out << "\\" << c;
-        else if (32 < n && n < 127) {
+        else if (32 <= n && n < 127 && n != '\"' && n != ' '
+            && n != '\\' && n != '\'' && n != '?' && n != '.' && n != '(' && n != ')' && n != '[' && n != ']'
+            && n != '{' && n != '}' && n != '&') {
             if (html_encode) {
                 if (c == '<')
                     out << "&lt;";
                 else if (c == '>')
                     out << "&gt;";
-                else if (c == '&')
-                    out << "&amp;";
-                else if (c == '\"')
-                    out << "&quot;";
+                //else if (c == '&')
+                //    out << "&amp;";
+                //else if (c == '\"')
+                //    out << "&quot;";
                 else
-                    out << "\\x" << std::hex << n;
+                    //out << "\\x" << std::hex << n;
+                    out << c;
             }
             else
                 out << c;
@@ -1118,91 +1226,193 @@ std::ostream& seq_util::rex::pp::seq_unit(std::ostream& out, expr* s) const {
             out << "\\x" << std::hex << n;
         else if (n <= 0xFFF)
             out << "\\u0" << std::hex << n;
-        else 
+        else
             out << "\\u" << std::hex << n;
     }
+    else if (re.u.str.is_nth_i(s, e, i)) {
+        print(out, e) << "[";
+        print(out, i) << "]";
+    }
+    else if (re.u.str.is_length(s, e))
+        print(out << "|", e) << "|";    
     else
-        out << "{" << mk_pp(s, re.m) << "}";
-    return out;
+        return false;
+    return true;
 }
 
 /*
-  Pretty prints the regex r into the out stream
+  Pretty prints the regex r into the ostream out
 */
-std::ostream& seq_util::rex::pp::display(std::ostream& out) const {
+std::ostream& seq_util::rex::pp::print(std::ostream& out, expr* e) const {
     expr* r1 = nullptr, * r2 = nullptr, * s = nullptr, * s2 = nullptr;
     unsigned lo = 0, hi = 0;
-    if (re.is_full_char(e))
-        return out << ".";
+    arith_util a(re.m);
+    rational v;
+    if (!e)
+        out << "null";
+    else if (print_unit(out, e))
+        ;
+    else if (print_seq(out, e))
+        ;
+    else if (re.is_full_char(e))
+        out << ".";
     else if (re.is_full_seq(e))
-        return out << ".*";
+        out << ".*";
     else if (re.is_to_re(e, s))
-        return compact_helper_seq(out, s);
-    else if (re.is_range(e, s, s2)) 
-        return compact_helper_range(out, s, s2);
+        print(out, s);
+    else if (re.is_range(e, s, s2))
+        print_range(out, s, s2);
     else if (re.is_epsilon(e))
-        return out << "()";
+        // &#x03B5; = epsilon
+        out << (html_encode ? "&#x03B5;" : "()");
     else if (re.is_empty(e))
-        return out << "[]";
-    else if (re.is_concat(e, r1, r2)) 
-        return out << pp(re, r1) << pp(re, r2);
-    else if (re.is_union(e, r1, r2)) 
-        return out << pp(re, r1) << "|" << pp(re, r2);
-    else if (re.is_intersection(e, r1, r2)) 
-        return out << "(" << pp(re, r1) << (html_encode ? ")&amp;(": ")&(" ) << pp(re, r2) << ")";
+        // &#x2205; = emptyset
+        out << (html_encode ? "&#x2205;" : "[]");
+    else if (re.is_concat(e, r1, r2)) {
+        print(out, r1);
+        print(out, r2);
+    }
+    else if (re.is_antimorov_union(e, r1, r2) || re.is_union(e, r1, r2)) {
+        out << "(";
+        print(out, r1);
+        out << (html_encode ? "&#x22C3;" : "|");
+        print(out, r2);
+        out << ")";
+    }
+    else if (re.is_intersection(e, r1, r2)) {
+        out << "(";
+        print(out, r1);
+        out << (html_encode ? "&#x22C2;" : "&");
+        print(out, r2);
+        out << ")";
+    }
     else if (re.is_complement(e, r1)) {
+        out << "~";
         if (can_skip_parenth(r1))
-            return out << "~" << pp(re, r1);
-        else 
-            return out << "~(" << pp(re, r1) << ")";
+            print(out, r1);
+        else {
+            out << "(";
+            print(out, r1);
+            out << ")";
+        }
     }
     else if (re.is_plus(e, r1)) {
-        if (can_skip_parenth(r1)) 
-            return out << pp(re, r1) << "+";
-        else 
-            return out << "(" << pp(re, r1) << ")+";
+        if (can_skip_parenth(r1)) {
+            print(out, r1);
+            out << "+";
+        }
+        else {
+            out << "(";
+            print(out, r1);
+            out << ")+";
+        }
     }
     else if (re.is_star(e, r1)) {
-        if (can_skip_parenth(r1))
-            return out << pp(re, r1) << "*";
-        else
-            return out << "(" << pp(re, r1) << ")*";
+        if (can_skip_parenth(r1)) {
+            print(out, r1);
+            out << "*";
+        }
+        else {
+            out << "(";
+            print(out, r1);
+            out << ")*";
+        }
     }
     else if (re.is_loop(e, r1, lo)) {
-        if (can_skip_parenth(r1))
-            return out << pp(re, r1) << "{" << lo << ",}";
-        else 
-            return out << "(" << pp(re, r1) << "){" << lo << ",}";
+        if (can_skip_parenth(r1)) 
+            print(out, r1) << "{" << lo << ",}";
+        else {
+            out << "(";
+            print(out, r1);
+            out << "){" << lo << ",}";
+        }
     }
     else if (re.is_loop(e, r1, lo, hi)) {
         if (can_skip_parenth(r1)) {
+            print(out, r1);
             if (lo == hi)
-                return out << pp(re, r1) << "{" << lo << "}";
-            else 
-                return out << pp(re, r1) << "{" << lo << "," << hi << "}";
+                out << "{" << lo << "}";
+            else
+                out << "{" << lo << "," << hi << "}";
         }
         else {
+            out << "(";
+            print(out, r1);
             if (lo == hi)
-                return out << "(" << pp(re, r1) << "){" << lo << "}";
+                out << "){" << lo << "}";
             else
-                return out << "(" << pp(re, r1) << "){" << lo << "," << hi << "}";
+                out << "){" << lo << "," << hi << "}";
         }
     }
-    else if (re.is_diff(e, r1, r2)) 
-        return out << "(" << pp(re, r1) << ")\\(" << pp(re, r2) << ")";
-    else if (re.m.is_ite(e, s, r1, r2)) 
-        return out << "if(" << mk_pp(s, re.m) << "," << pp(re, r1) << "," << pp(re, r2) << ")";
+    else if (re.is_diff(e, r1, r2)) {
+        out << "(";
+        print(out, r1);
+        out << ")\\(";
+        print(out, r2);
+        out << ")";
+    }
+    else if (re.m.is_ite(e, s, r1, r2)) {
+        out << (html_encode ? "(&#x1D422;&#x1D41F; " : "(if ");
+        print(out, s);
+        out << (html_encode ? " &#x1D42D;&#x1D5F5;&#x1D41E;&#x1D427; " : " then ");
+        print(out, r1);
+        out << (html_encode ? " &#x1D41E;&#x1D425;&#x1D600;&#x1D41E; " : " else ");
+        print(out, r2);
+        out << ")";
+    }
     else if (re.is_opt(e, r1)) {
         if (can_skip_parenth(r1)) 
-            return out << pp(re, r1) << "?";
-        else 
-            return out << "(" << pp(re, r1) << ")?";
+            print(out, r1) << "?";
+        else {
+            out << "(";
+            print(out, r1);
+            out << ")?";
+        }
+    }
+    else if (re.is_reverse(e, r1)) {
+        out << "(reverse ";
+        print(out, r1);
+        out << ")";
+    }
+    else if (re.m.is_eq(e, r1, r2)) {
+        out << "(";
+        print(out, r1);
+        out << " = ";
+        print(out, r2);
+        out << ")";
+    }
+    else if (re.m.is_not(e, r1)) {
+        out << "!";
+        print(out, r1);
+    }
+    else if (a.is_add(e, s, s2) && a.is_numeral(s, v) && v < 0) 
+        print(out, s2) << " - " << -v;
+    else if (a.is_add(e, s, s2) && a.is_numeral(s2, v) && v < 0)
+        print(out, s) << " - " << -v;
+    else if (a.is_add(e, s, s2))
+        print(out, s) << " + ", print(out, s2);
+    else if (a.is_sub(e, s, s2) && a.is_numeral(s2, v) && v > 0)
+        print(out, s) << " - " << v;
+    else if (a.is_le(e, s, s2))
+        print(out << "(", s) << " <= ", print(out, s2) << ")";
+    else if (re.m.is_value(e))
+        out << mk_pp(e, re.m);
+    else if (is_app(e) && to_app(e)->get_num_args() == 0)
+        out << mk_pp(e, re.m);
+    else if (is_app(e)) {
+        out << "(" << to_app(e)->get_decl()->get_name();
+        for (expr* arg : *to_app(e))
+            print(out << " ", arg);
+        out << ")";
     }
-    else if (re.is_reverse(e, r1)) 
-        return out << "reverse(" << pp(re, r1) << ")";
     else
-        // Else: derivative or is_of_pred
-        return out << "{" << mk_pp(e, re.m) << "}";
+        // for all remaining cases use the default pretty printer
+        out << mk_pp(e, re.m);
+    return out;
+}
+
+std::ostream& seq_util::rex::pp::display(std::ostream& out) const {
+    return print(out, ex);
 }
 
 /*
@@ -1210,7 +1420,16 @@ std::ostream& seq_util::rex::pp::display(std::ostream& out) const {
 */
 std::string seq_util::rex::to_str(expr* r) const {
     std::ostringstream out;
-    out << pp(u.re, r);
+    pp(u.re, r, false).display(out);
+    return out.str();
+}
+
+/*
+  Pretty prints the regex r into the output string that is htmlencoded 
+*/
+std::string seq_util::rex::to_strh(expr* r) const {
+    std::ostringstream out;
+    pp(u.re, r, true).display(out);
     return out.str();
 }
 
@@ -1256,7 +1475,7 @@ seq_util::rex::info seq_util::rex::get_info_rec(expr* e) const {
     else 
         result = mk_info_rec(to_app(e));
     m_infos.setx(e->get_id(), result, invalid_info);
-    STRACE("re_info", tout << "compute_info(" << pp(u.re, e) << ")=" << result << std::endl;);
+    STRACE("re_info", tout << "compute_info(" << pp(u.re, e, false) << ")=" << result << std::endl;);
     return result;
 }
 
@@ -1484,7 +1703,13 @@ seq_util::rex::info seq_util::rex::info::orelse(seq_util::rex::info const& i) co
             // unsigned ite_min_length = std::min(min_length, i.min_length);
             // lbool ite_nullable = (nullable == i.nullable ? nullable : l_undef);
             // TBD: whether ite is interpreted or not depends on whether the condition is interpreted and both branches are interpreted
-            return info(false, false, false, false, normalized && i.normalized, monadic && i.monadic, singleton && i.singleton, nullable, std::min(min_length, i.min_length), std::max(star_height, i.star_height));
+            return info(false, false, false, false, 
+                normalized && i.normalized, 
+                monadic && i.monadic, 
+                singleton && i.singleton, 
+                ((nullable == l_true && i.nullable == l_true) ? l_true : ((nullable == l_false && i.nullable == l_false) ? l_false : l_undef)),
+                std::min(min_length, i.min_length), 
+                std::max(star_height, i.star_height));
         }
         else
             return i;
