@@ -246,7 +246,7 @@ namespace qe {
         app_ref_vector avars(m); 
         bool_vector seen;
         arith_util a(m);
-        for (expr* e : subterms(lits)) {
+        for (expr* e : subterms::ground(lits)) {
             if ((m.is_eq(e) && a.is_int_real(to_app(e)->get_arg(0))) || a.is_arith_expr(e)) {
                 for (expr* arg : *to_app(e)) {
                     unsigned id = arg->get_id();
@@ -368,7 +368,7 @@ namespace qe {
     void uflia_mbi::add_arith_dcert(model& mdl, expr_ref_vector& lits) {
         obj_map<func_decl, ptr_vector<app>> apps;
         arith_util a(m);
-        for (expr* e : subterms(lits)) {
+        for (expr* e : subterms::ground(lits)) {
             if (a.is_int_real(e) && is_uninterp(e) && to_app(e)->get_num_args() > 0) {
                 func_decl* f = to_app(e)->get_decl();
                 apps.insert_if_not_there(f, ptr_vector<app>()).push_back(to_app(e));
