@@ -450,7 +450,7 @@ namespace arith {
         lp::lar_solver& lp() { return *m_solver; }
         lp::lar_solver const& lp() const { return *m_solver; }
         bool is_equal(theory_var x, theory_var y) const;
-        void add_eq(lpvar u, lpvar v, lp::explanation const& e);
+        bool add_eq(lpvar u, lpvar v, lp::explanation const& e, bool is_fixed);
         void consume(rational const& v, lp::constraint_index j);
         bool bound_is_interesting(unsigned vi, lp::lconstraint_kind kind, const rational& bval) const;
     };
