@@ -75,10 +75,11 @@ namespace array {
     }
 
     void solver::internalize_lambda(euf::enode* n) {
-        set_prop_upward(n);
-        if (!a.is_store(n->get_expr()))
-            push_axiom(default_axiom(n));
-        add_lambda(n->get_th_var(get_id()), n);
+        SASSERT(is_lambda(n->get_expr()) || a.is_const(n->get_expr()) || a.is_as_array(n->get_expr()));
+        theory_var v = n->get_th_var(get_id());
+        push_axiom(default_axiom(n));
+        add_lambda(v, n);
+        set_prop_upward(v);
     }
 
     void solver::internalize_select(euf::enode* n) {
@@ -119,8 +120,8 @@ namespace array {
         SASSERT(!n || !n->is_attached_to(get_id()));
         if (!n) 
             n = mk_enode(e, false);
-        SASSERT(!n->is_attached_to(get_id()));
-        mk_var(n);
+        if (!n->is_attached_to(get_id())) 
+            mk_var(n);
         for (auto* arg : euf::enode_args(n))
             ensure_var(arg);  
         switch (a->get_decl_kind()) {
@@ -177,6 +178,8 @@ namespace array {
         auto set_index = [&](euf::enode* arg) { if (arg->get_root() == r) is_index = true; };
         auto set_value = [&](euf::enode* arg) { if (arg->get_root() == r) is_value = true; };
 
+        if (a.is_ext(n->get_expr()))
+            return true;
         for (euf::enode* parent : euf::enode_parents(r)) {
             app* p = parent->get_app();
             unsigned num_args = parent->num_args();
@@ -193,7 +196,7 @@ namespace array {
             }
             else if (a.is_const(p)) {
                 set_value(parent->get_arg(0));
-            }
+            }            
             if (is_array + is_index + is_value > 1)
                 return true;
         }
