@@ -22,6 +22,7 @@ Author:
 #include "ast/euf/euf_enode.h"
 #include "sat/smt/euf_solver.h"
 
+
 namespace q {
 
     struct lit {
@@ -29,39 +30,25 @@ namespace q {
         expr_ref rhs;
         bool     sign;
         lit(expr_ref const& lhs, expr_ref const& rhs, bool sign):
-            lhs(lhs), rhs(rhs), sign(sign) {}
+            lhs(lhs), rhs(rhs), sign(sign) {
+            SASSERT(!rhs.m().is_false(rhs) || !sign);
+        }
         std::ostream& display(std::ostream& out) const;
     };
 
-    struct binding : public dll_base<binding> {
-        app*               m_pattern;
-        unsigned           m_max_generation;
-        unsigned           m_min_top_generation;
-        unsigned           m_max_top_generation;
-        euf::enode*        m_nodes[0];
-
-        binding(app* pat, unsigned max_generation, unsigned min_top, unsigned max_top):
-            m_pattern(pat),
-            m_max_generation(max_generation),
-            m_min_top_generation(min_top),
-            m_max_top_generation(max_top) {}
-
-        euf::enode* const* nodes() { return m_nodes; }
-
-        euf::enode* operator[](unsigned i) const { return m_nodes[i]; }
-
-        std::ostream& display(euf::solver& ctx, unsigned num_nodes, std::ostream& out) const;
-    };
+    struct binding;
 
     struct clause {
         unsigned            m_index;
         vector<lit>         m_lits;
         quantifier_ref      m_q;
-        sat::literal        m_literal;
+        unsigned            m_watch = 0;
+        sat::literal        m_literal = sat::null_literal;
         q::quantifier_stat* m_stat = nullptr;
-        binding*            m_bindings = nullptr;
+        binding* m_bindings = nullptr;
 
-        clause(ast_manager& m, unsigned idx): m_index(idx), m_q(m) {}
+
+        clause(ast_manager& m, unsigned idx) : m_index(idx), m_q(m) {}
 
         std::ostream& display(euf::solver& ctx, std::ostream& out) const;
         lit const& operator[](unsigned i) const { return m_lits[i]; }
@@ -72,13 +59,79 @@ namespace q {
         quantifier* q() const { return m_q; }
     };
 
+
+    struct binding : public dll_base<binding> {
+        clause*            c;
+        app*               m_pattern;
+        unsigned           m_max_generation;
+        unsigned           m_min_top_generation;
+        unsigned           m_max_top_generation;
+        euf::enode*        m_nodes[0];
+
+        binding(clause& c, app* pat, unsigned max_generation, unsigned min_top, unsigned max_top):
+            c(&c),
+            m_pattern(pat),
+            m_max_generation(max_generation),
+            m_min_top_generation(min_top),
+            m_max_top_generation(max_top) {}
+
+        euf::enode* const* nodes() { return m_nodes; }
+
+        euf::enode* operator[](unsigned i) const { return m_nodes[i]; }
+
+        std::ostream& display(euf::solver& ctx, std::ostream& out) const;
+
+        unsigned size() const { return c->num_decls(); }
+        
+        quantifier* q() const { return c->m_q; }
+
+        bool eq(binding const& other) const {
+            if (q() != other.q())
+                return false;
+            for (unsigned i = size(); i-- > 0; )
+                if ((*this)[i] != other[i])
+                    return false;
+            return true;
+        }
+    };
+
+    struct binding_khasher {
+        unsigned operator()(binding const* f) const { return f->q()->get_id(); }
+    };
+
+    struct binding_chasher {
+        unsigned operator()(binding const* f, unsigned idx) const { return f->m_nodes[idx]->hash(); }
+    };
+
+    struct binding_hash_proc {
+        unsigned operator()(binding const* f) const {
+            return get_composite_hash<binding*, binding_khasher, binding_chasher>(const_cast<binding*>(f), f->size());
+        }
+    };
+
+    struct binding_eq_proc {
+        bool operator()(binding const* a, binding const* b) const { return a->eq(*b); }
+    };
+
+    typedef ptr_hashtable<binding, binding_hash_proc, binding_eq_proc> bindings;
+
+    inline std::ostream& operator<<(std::ostream& out, binding const& f) {
+        out << "[fp " << f.q()->get_id() << ":";
+        for (unsigned i = 0; i < f.size(); ++i)
+            out << " " << f[i]->get_expr_id();
+        return out << "]";
+    }
+
+
     struct justification {
         expr*     m_lhs, *m_rhs;
         bool      m_sign;
+        unsigned  m_num_ev;
+        euf::enode_pair* m_evidence;
         clause&   m_clause;
         euf::enode* const* m_binding;
-        justification(lit const& l, clause& c, euf::enode* const* b):
-            m_lhs(l.lhs), m_rhs(l.rhs), m_sign(l.sign), m_clause(c), m_binding(b) {}
+        justification(lit const& l, clause& c, euf::enode* const* b, unsigned n, euf::enode_pair* ev):
+            m_lhs(l.lhs), m_rhs(l.rhs), m_sign(l.sign), m_num_ev(n), m_evidence(ev), m_clause(c), m_binding(b) {}
         sat::ext_constraint_idx to_index() const { 
             return sat::constraint_base::mem2base(this); 
         }
