@@ -19,11 +19,11 @@ Author:
 #include "util/nat_set.h"
 #include "ast/quantifier_stat.h"
 #include "ast/pattern/pattern_inference.h"
+#include "ast/normal_forms/nnf.h"
 #include "solver/solver.h"
 #include "sat/smt/sat_th.h"
 #include "sat/smt/q_mam.h"
 #include "sat/smt/q_clause.h"
-#include "sat/smt/q_fingerprint.h"
 #include "sat/smt/q_queue.h"
 #include "sat/smt/q_eval.h"
 
@@ -41,6 +41,7 @@ namespace q {
             unsigned m_num_propagations;
             unsigned m_num_conflicts;
             unsigned m_num_redundant;
+            unsigned m_num_delayed_bindings;
             
             stats() { reset(); }
 
@@ -49,6 +50,13 @@ namespace q {
             }
         };
 
+        struct prop {
+            bool is_conflict;
+            unsigned idx;
+            sat::ext_justification_idx j;
+            prop(bool is_conflict, unsigned idx, sat::ext_justification_idx j) : is_conflict(is_conflict), idx(idx), j(j) {}
+        };
+
         struct remove_binding;
         struct insert_binding;
         struct pop_clause;
@@ -61,10 +69,11 @@ namespace q {
         ast_manager&                  m;
         eval                          m_eval;
         quantifier_stat_gen           m_qstat_gen;
-        fingerprints                  m_fingerprints;
+        bindings                      m_bindings;
         scoped_ptr<binding>           m_tmp_binding;
-        unsigned                      m_tmp_binding_capacity { 0 };
+        unsigned                      m_tmp_binding_capacity = 0;
         queue                         m_inst_queue;
+        svector<prop>                 m_prop_queue;
         pattern_inference_rw          m_infer_patterns;
         scoped_ptr<q::mam>            m_mam, m_lazy_mam;
         ptr_vector<clause>            m_clauses;
@@ -72,24 +81,25 @@ namespace q {
         vector<unsigned_vector>       m_watch;     // expr_id -> clause-index*
         stats                         m_stats;
         expr_fast_mark1               m_mark;
-        unsigned                      m_generation_propagation_threshold{ 3 };
+        unsigned                      m_generation_propagation_threshold = 3;
         ptr_vector<app>               m_ground;
-        bool                          m_in_queue_set{ false };
+        bool                          m_in_queue_set = false;
         nat_set                       m_node_in_queue;
         nat_set                       m_clause_in_queue;
-        unsigned                      m_qhead { 0 };
+        unsigned                      m_qhead = 0;
         unsigned_vector               m_clause_queue;
+        euf::enode_pair_vector        m_evidence;
 
-        binding* alloc_binding(unsigned n, app* pat, unsigned max_generation, unsigned min_top, unsigned max_top);
-        euf::enode* const* alloc_binding(clause& c, euf::enode* const* _binding);
-        void add_binding(clause& c, app* pat, euf::enode* const* _binding, unsigned max_generation, unsigned min_top, unsigned max_top);
-
+        euf::enode* const* copy_nodes(clause& c, euf::enode* const* _binding);
+        binding* tmp_binding(clause& c, app* pat, euf::enode* const* _binding);
+        binding* alloc_binding(clause& c, app* pat, euf::enode* const* _binding, unsigned max_generation, unsigned min_top, unsigned max_top);
+        
         sat::ext_justification_idx mk_justification(unsigned idx, clause& c, euf::enode* const* b);
 
         void ensure_ground_enodes(expr* e);
         void ensure_ground_enodes(clause const& c);
 
-        void instantiate(binding& b, clause& c);
+        void instantiate(binding& b);
         sat::literal instantiate(clause& c, euf::enode* const* binding, lit const& l);
 
         // register as callback into egraph.
@@ -103,9 +113,19 @@ namespace q {
 
         void attach_ground_pattern_terms(expr* pat);
         clause* clausify(quantifier* q);
+        lit clausify_literal(expr* arg);
 
-        fingerprint* add_fingerprint(clause& c, binding& b, unsigned max_generation);
-        void set_tmp_binding(fingerprint& fp);
+        bool flush_prop_queue();
+        void propagate(bool is_conflict, unsigned idx, sat::ext_justification_idx j_idx);
+
+        bool propagate(bool flush);
+
+        expr_ref_vector m_new_defs;
+        proof_ref_vector m_new_proofs;
+        defined_names    m_dn;
+        nnf              m_nnf;
+        
+        quantifier_ref nnf_skolem(quantifier* q);
 
     public:
         
@@ -113,9 +133,8 @@ namespace q {
             
         bool operator()();
 
-        bool propagate(bool flush);
+        bool unit_propagate();        
 
-        void init_search();
 
         void add(quantifier* q);
 
@@ -127,7 +146,7 @@ namespace q {
         void on_binding(quantifier* q, app* pat, euf::enode* const* binding, unsigned max_generation, unsigned min_gen, unsigned max_gen);
 
         // callbacks from queue
-        lbool evaluate(euf::enode* const* binding, clause& c) { return m_eval(binding, c); }
+        lbool evaluate(euf::enode* const* binding, clause& c) { m_evidence.reset();  return m_eval(binding, c, m_evidence); }
 
         void add_instantiation(clause& c, binding& b, sat::literal lit);
 
