@@ -124,6 +124,12 @@ namespace smt {
                 }
             }
 
+            scoped_trace_stream(theory& th, literal_buffer const& lits) : m(th.get_manager()) {
+                if (m.has_trace_stream()) {
+                    th.log_axiom_instantiation(lits);
+                }
+            }
+
             scoped_trace_stream(theory& th, literal lit): m(th.get_manager()) {
                 if (m.has_trace_stream()) {
                     literal_vector lits;
@@ -447,9 +453,6 @@ namespace smt {
         
         std::ostream& display_flat_app(std::ostream & out, app * n) const;
         
-        std::ostream& display_var_def(std::ostream & out, theory_var v) const { return display_app(out, get_enode(v)->get_expr()); }
-        
-        std::ostream& display_var_flat_def(std::ostream & out, theory_var v) const { return display_flat_app(out, get_enode(v)->get_expr());  }
 
     protected:
         void log_axiom_instantiation(app * r, unsigned axiom_id = UINT_MAX, unsigned num_bindings = 0, 
@@ -464,6 +467,8 @@ namespace smt {
 
         void log_axiom_instantiation(literal_vector const& ls);
 
+        void log_axiom_instantiation(literal_buffer const& ls);
+
         void log_axiom_instantiation(app * r, unsigned num_blamed_enodes, enode ** blamed_enodes) {
             vector<std::tuple<enode *, enode *>> used_enodes;
             for (unsigned i = 0; i < num_blamed_enodes; ++i) {
@@ -473,6 +478,7 @@ namespace smt {
         }
 
         void log_axiom_unit(app* r) {
+            expr_ref _r(r, m);
             log_axiom_instantiation(r);
             m.trace_stream() << "[end-of-instance]\n";
         }
