@@ -213,13 +213,13 @@ void asserted_formulas::force_push() {
 }
 
 void asserted_formulas::pop_scope(unsigned num_scopes) {
-    if (m_lazy_scopes > 0) {
-        unsigned n = std::min(num_scopes, m_lazy_scopes);
-        m_lazy_scopes -= n;
-        num_scopes -= n;
-        if (num_scopes == 0)
-            return;
+    if (num_scopes <= m_lazy_scopes) {
+        m_lazy_scopes -= num_scopes;
+        return;
     }
+    num_scopes -= m_lazy_scopes;
+    m_lazy_scopes = 0;
+    
     TRACE("asserted_formulas_scopes", tout << "before pop " << num_scopes << " of " << m_scopes.size() << "\n";);
     m_bv_sharing.pop_scope(num_scopes);
     m_macro_manager.pop_scope(num_scopes);
@@ -424,7 +424,8 @@ void asserted_formulas::apply_quasi_macros() {
     TRACE("before_quasi_macros", display(tout););
     vector<justified_expr> new_fmls;
     quasi_macros proc(m, m_macro_manager);
-    while (proc(m_formulas.size() - m_qhead,
+    while (m_qhead == 0 && 
+           proc(m_formulas.size() - m_qhead,
                 m_formulas.data() + m_qhead,
                 new_fmls)) {
         swap_asserted_formulas(new_fmls);
@@ -531,14 +532,14 @@ void asserted_formulas::propagate_values() {
     flush_cache();
 
     unsigned num_prop = 0;
-    unsigned delta_prop = m_formulas.size();
-    while (!inconsistent() && m_formulas.size()/20 < delta_prop) {
+    unsigned sz = m_formulas.size();
+    unsigned delta_prop = sz;
+    while (!inconsistent() && sz/20 < delta_prop) {
         m_expr2depth.reset();
         m_scoped_substitution.push();
         unsigned prop = num_prop;
         TRACE("propagate_values", display(tout << "before:\n"););
         unsigned i  = m_qhead;
-        unsigned sz = m_formulas.size();
         for (; i < sz; i++) {
             prop += propagate_values(i);
         }
@@ -557,6 +558,9 @@ void asserted_formulas::propagate_values() {
         TRACE("propagate_values", tout << "after:\n"; display(tout););
         delta_prop = prop - num_prop;
         num_prop = prop;
+        if (sz <= m_formulas.size())
+            break;
+        sz = m_formulas.size();
     }
     TRACE("asserted_formulas", tout << num_prop << "\n";);
     if (num_prop > 0)
