@@ -503,7 +503,10 @@ private:
         cube.append(s.split_cubes(1));
         SASSERT(cube.size() <= 1);
         IF_VERBOSE(2, verbose_stream() << "(tactic.parallel :split-cube " << cube.size() << ")\n";);
-        if (!s.cubes().empty()) m_queue.add_task(s.clone());
+        {
+            // std::lock_guard<std::mutex> lock(m_mutex);
+            if (!s.cubes().empty()) m_queue.add_task(s.clone());
+        }
         if (!cube.empty()) {
             s.assert_cube(cube.get(0).cube());
             vars.reset();
@@ -611,8 +614,12 @@ private:
     void spawn_cubes(solver_state& s, unsigned width, vector<cube_var>& cubes) {
         if (cubes.empty()) return;
         add_branches(cubes.size());
-        s.set_cubes(cubes);
-        solver_state* s1 = s.clone();
+        s.set_cubes(cubes);        
+        solver_state* s1 = nullptr;
+        {
+            // std::lock_guard<std::mutex> lock(m_mutex);
+            s1 = s.clone();
+        }
         s1->inc_width(width);
         m_queue.add_task(s1);
     }
