@@ -93,7 +93,8 @@ class fm_tactic : public tactic {
                     expr * lhs = to_app(l)->get_arg(0);
                     expr * rhs = to_app(l)->get_arg(1);
                     rational c;
-                    u.is_numeral(rhs, c);
+                    if (!u.is_numeral(rhs, c))
+                        return NONE;
                     if (neg)
                         c.neg();
                     unsigned num_mons;
@@ -112,7 +113,8 @@ class fm_tactic : public tactic {
                         expr * xi;
                         rational ai_val;
                         if (u.is_mul(monomial, ai, xi)) {
-                            u.is_numeral(ai, ai_val);
+                            if (!u.is_numeral(ai, ai_val))
+                                return NONE;
                         }
                         else {
                             xi     = monomial;
@@ -120,7 +122,8 @@ class fm_tactic : public tactic {
                         }
                         if (u.is_to_real(xi))
                             xi = to_app(xi)->get_arg(0);
-                        SASSERT(is_uninterp_const(xi));
+                        if (!is_uninterp_const(xi))
+                            return NONE;
                         if (x == to_app(xi)->get_decl()) {
                             a_val = ai_val;
                             if (neg)
@@ -129,9 +132,9 @@ class fm_tactic : public tactic {
                         else {
                             expr_ref val(m);
                             val = ev(monomial);
-                            SASSERT(u.is_numeral(val));
                             rational tmp;
-                            u.is_numeral(val, tmp);
+                            if (!u.is_numeral(val, tmp))
+                                return NONE;
                             if (neg)
                                 tmp.neg();
                             c -= tmp;
@@ -162,11 +165,9 @@ class fm_tactic : public tactic {
         fm_model_converter(ast_manager & _m):m(_m) {}
 
         ~fm_model_converter() override {
-            m.dec_array_ref(m_xs.size(), m_xs.data());
-            vector<clauses>::iterator it  = m_clauses.begin();
-            vector<clauses>::iterator end = m_clauses.end();
-            for (; it != end; ++it)
-                m.dec_array_ref(it->size(), it->data());
+            m.dec_array_ref(m_xs.size(), m_xs.data());     
+            for (auto& c : m_clauses)
+                m.dec_array_ref(c.size(), c.data());
         }
         
         void insert(func_decl * x, clauses & c) {
@@ -177,6 +178,7 @@ class fm_tactic : public tactic {
             m_clauses.back().swap(c);
         }
 
+
         void get_units(obj_map<expr, bool>& units) override { units.reset(); }
 
         void operator()(model_ref & md) override {
@@ -252,11 +254,8 @@ class fm_tactic : public tactic {
             for (unsigned i = 0; i < sz; i++) {
                 out << "\n(" << m_xs[i]->get_name();
                 clauses const & cs = m_clauses[i];
-                clauses::const_iterator it  = cs.begin();
-                clauses::const_iterator end = cs.end();
-                for (; it != end; ++it) {
-                    out << "\n  " << mk_ismt2_pp(*it, m, 2);
-                }
+                for (auto& c : cs)
+                    out << "\n  " << mk_ismt2_pp(c, m, 2);                
                 out << ")";
             }
             out << ")\n";
@@ -274,10 +273,8 @@ class fm_tactic : public tactic {
                 clauses const & cs = m_clauses[i];
                 res->m_clauses.push_back(clauses());
                 clauses & new_cs = res->m_clauses.back();
-                clauses::const_iterator it  = cs.begin();
-                clauses::const_iterator end = cs.end();
-                for (; it != end; ++it) {
-                    app * new_c = translator(*it);
+                for (auto& c : cs) {
+                    app * new_c = translator(c);
                     to_m.inc_ref(new_c);
                     new_cs.push_back(new_c);
                 }
@@ -1598,7 +1595,7 @@ class fm_tactic : public tactic {
                 report_tactic_progress(":fm-cost", m_counter);
                 if (!m_inconsistent) {
                     copy_remaining();
-                    m_new_goal->add(concat(g->mc(), m_mc.get()));
+                    m_new_goal->add(m_mc.get());
                 }
             }
             reset_constraints();
